-- This module provides some combinators for equational reasoning
-- about words. It also provides some basic lemmas about monoids and
-- groups.

open import Lib.Proposition
open import Lib.Equality as Eq using (_==_)
open import Lib.Nat
open import Lib.List
open import Lib.Maybe
open import Lib.MaybeEq

open import Lib.Word

module Lib.Word-Lemmas where

lemma-wmap : ∀ {A} {Γ : Context A} -> (f : A -> A) -> (∀ (a : A) -> Γ ⊢ gen a === gen (f a)) -> ∀ (w : Word A) -> Γ ⊢ w === wmap f w
lemma-wmap f pf (gen x) = pf x
lemma-wmap f pf ε = refl
lemma-wmap f pf (w • w₁) = cong (lemma-wmap f pf w) (lemma-wmap f pf w₁)


isL : ∀ {A B} -> (w : Word (A ∪ B)) -> Maybe (∃ λ (w' : Word A) -> ⌞ w' == w)
isL (gen (⌜ x)) = Just ((gen x) , Eq.refl)
isL (gen (x ⌝)) = Nothing
isL ε = Just (ε , Eq.refl)
isL (w • v) with isL w | isL v
... | Just (w' , pr1) | Just (v' , pr2) = Just ((w' • v') , Eq.cong _•_ pr1 pr2)
... | _ | _ = Nothing


isR : ∀ {A B} -> (w : Word (A ∪ B)) -> Maybe (∃ λ (w' : Word B) -> w' ⌟ == w)
isR (gen (⌜ x)) = Nothing
isR (gen (x ⌝)) = Just ((gen x) , Eq.refl)
isR ε = Just (ε , Eq.refl)
isR (w • v) with isR w | isR v
... | Just (w' , pr1) | Just (v' , pr2) = Just ((w' • v') , Eq.cong _•_ pr1 pr2)
... | _ | _ = Nothing


-- ----------------------------------------------------------------------
-- * Combinators for equational reasoning

module Monoid-Equational where

  -- The following definitions provide a convenient syntax for writing
  -- equational proofs in monoids. To prove Γ ⊢ a0 === an, we can write:
  -- 
  -- equational a0
  --         by reason1
  --     equals a1
  --         by reason2
  --     equals a2
  --        ...
  --         by reasonn
  --     equals an
  --
  -- Note that the syntax is the same as that provided in Equality.agda,
  -- so depending on which type of equation one needs to reason about,
  -- only one or the other module should be opened.
  
  infix 2 equational_
  infixl 1 by-equals
  infixl 4 _reversed
  infixl 3 left_
  infixl 3 right_

  equational_ : ∀ {X : Set} {Γ : Context X} -> (w : Word X) -> Γ ⊢ w === w
  equational_ w = refl
  
  by-equals : ∀ {X : Set} {Γ : Context X} {w v : Word X} -> Γ ⊢ w === v -> (u : Word X) -> Γ ⊢ v === u -> Γ ⊢ w === u
  by-equals E u reason = trans E reason
  
  syntax by-equals E u reason = E by reason equals u
  
  _reversed : ∀ {X : Set} {Γ : Context X} {w v : Word X} -> Γ ⊢ w === v -> Γ ⊢ v === w
  _reversed = symm
  
  left_ : ∀ {X : Set} {Γ : Context X} {w v u : Word X} -> Γ ⊢ w === v -> Γ ⊢ w • u === v • u
  left_ E = cong E refl
  
  right_ : ∀ {X : Set} {Γ : Context X} {w v u : Word X} -> Γ ⊢ w === v -> Γ ⊢ u • w === u • v
  right_ E = cong refl E
  
  -- A version of reflexivity that can be used as an introduction rule.
  refl' : ∀ {X : Set} {Γ : Context X} {w v : Word X} -> w == v -> Γ ⊢ w === v
  refl' Eq.refl = refl
  
  -- A synonym for refl, so that we can say "by definition".
  definition : ∀ {X : Set} {Γ : Context X} {w} -> Γ ⊢ w === w
  definition = refl


  infixl 4 _=W?_
  _=W?_ : ∀ {A} -> {{MaybeEq A}} -> (xs ys : Word A) -> Maybe (xs == ys) 
  gen x =W? gen x₁ with x =m? x₁
  ... | Just x₂ = Just (Eq.context gen x₂)
  ... | Nothing = Nothing
  gen x =W? ε = Nothing
  gen x =W? y • y₁ = Nothing
  ε =W? gen x = Nothing
  ε =W? ε = Just Eq.refl
  ε =W? y • y₁ = Nothing
  x • x₁ =W? gen x₂ = Nothing
  x • x₁ =W? ε = Nothing
  x • x₁ =W? y • y₁ with x =W? y | x₁ =W? y₁
  ... | Just x₂ | Just x₃ = Just (Eq.cong _•_ x₂ x₃)
  ... | Just x₂ | Nothing = Nothing
  ... | Nothing | h2 = Nothing

  instance 
    MaybeEq-Word : ∀ {A} -> {{MaybeEq A}} -> MaybeEq (Word A)
    MaybeEq-Word ._=m?_ = _=W?_ 

  flatten-symm : ∀ {X : Set} {Γ : Context X} {l r} -> {{meq : MaybeEq X}} -> (n : Nat) -> Γ ⊢ l === r -> Γ ⊢ l === r
  flatten-symm zero d = d
  flatten-symm (succ n) (trans d1 d2) = trans (flatten-symm n d1) (flatten-symm n d2)
  flatten-symm (succ n) (cong d1 d2) = cong (flatten-symm n d1) (flatten-symm n d2)
  flatten-symm (succ n) (symm refl) = refl
  flatten-symm (succ n) (symm (symm d)) = flatten-symm n d
  flatten-symm (succ n) (symm (trans d d₁)) = trans (flatten-symm n (symm d₁)) (flatten-symm n (symm d))
  flatten-symm (succ n) (symm (cong d d₁)) = cong (flatten-symm n (symm d)) (flatten-symm n (symm d₁))
  flatten-symm (succ n) (symm d) = symm d
  flatten-symm (succ n) d = d

  flatten-cong : ∀ {X : Set} {Γ : Context X} {l r} -> {{meq : MaybeEq X}} -> (n : Nat) -> Γ ⊢ l === r -> Γ ⊢ l === r
  flatten-cong zero d = d
  flatten-cong (succ n) (trans d1 d2) = trans (flatten-cong n d1) (flatten-cong n d2)
  flatten-cong (succ n) (cong refl refl) = refl
  flatten-cong (succ n) (cong refl (cong refl d)) = trans (symm assoc) (trans (flatten-cong n (cong refl d)) assoc)
  flatten-cong (succ n) (cong (trans d0 d1) (trans d2 d3)) = trans (flatten-cong n (cong d0 d2)) (flatten-cong n (cong d1 d3))
  flatten-cong (succ n) (cong (trans d0 d1) d2) = trans (flatten-cong n (cong d0 d2)) (flatten-cong n (cong d1 refl))
  flatten-cong (succ n) (cong d1 (trans d2 d3)) = trans (flatten-cong n (cong d1 d2)) (flatten-cong n (cong refl d3))
  flatten-cong (succ n) (cong (cong d0 d1) (cong d2 d3)) = trans assoc (trans (flatten-cong n (cong d0 (cong d1 (cong d2 d3)))) (symm assoc))
  flatten-cong (succ n) (cong (cong d0 d1) d2) = trans assoc (trans (flatten-cong n (cong d0 (cong d1 d2))) (symm assoc))
  flatten-cong (succ n) (cong d1 (cong d2 d3)) with flatten-cong n (cong d2 d3)
  flatten-cong (succ n) (cong d1 (cong d2 d3)) | trans d4 d5 = trans (flatten-cong n (cong d1 d4)) (flatten-cong n (cong refl d5))
  flatten-cong (succ n) (cong d1 (cong d2 d3)) | cong d4 d5 with d4 | d1
  flatten-cong (succ n) (cong d1 (cong d2 d3)) | cong d4 d5 | refl | refl = trans (symm assoc) (trans (cong refl d5) assoc)
  flatten-cong (succ n) (cong d1 (cong d2 d3)) | cong d4 d5 | d4' | d1' = cong d1 (cong d4 d5)
  flatten-cong (succ n) (cong d1 (cong d2 d3)) | atom = cong d1 atom
  flatten-cong (succ n) (cong d1 d2) = cong d1 d2
  flatten-cong (succ n) d = d

  flatten-trans : ∀ {X : Set} {Γ : Context X} {l r} -> {{meq : MaybeEq X}} -> (n : Nat) -> Γ ⊢ l === r -> Γ ⊢ l === r
  flatten-trans zero d = d
  -- testing
  flatten-trans (succ n) (cong d1 d2) = cong (flatten-trans n d1) (flatten-trans n d2)
  flatten-trans (succ n) (trans refl d) = flatten-trans n d
  flatten-trans (succ n) (trans d refl) = flatten-trans n d
  flatten-trans (succ n) (trans assoc (trans (symm assoc) d)) = flatten-trans n d
  flatten-trans (succ n) (trans (symm assoc) (trans assoc d)) = flatten-trans n d
  flatten-trans (succ n) (trans (trans d1 d2) d3) = flatten-trans n (trans d1 (trans d2 d3))
  flatten-trans (succ n) (trans d1 d2) = trans d1 (flatten-trans n d2)
  flatten-trans (succ n) d = d


  shorten-der1 : ∀ {X : Set} {Γ : Context X} {v t} -> {{meq : MaybeEq X}} -> (w : Word X) -> Γ ⊢ v === t -> Maybe (Γ ⊢ w === t)
  shorten-der1 {v = v} {t = t} w (axiom x) with w =m? t
  ... | Just x₁ rewrite x₁ = Just refl
  ... | Nothing with w =m? v
  ... | Nothing = Nothing
  ... | Just p rewrite p = Just (axiom x)
  shorten-der1 {v = v} w refl with w =m? v
  ... | Just x rewrite x = Just refl
  ... | Nothing = Nothing
  shorten-der1 {v = v} {t = t} w (symm d) with w =m? t
  ... | Just x₁ rewrite x₁ = Just refl
  ... | Nothing with w =m? v
  ... | Nothing = Nothing
  ... | Just p rewrite p = Just ((symm d))
  shorten-der1 w (trans {v₁} {v = v} d d₁) with w =m? v₁
  ... | Nothing = shorten-der1 w d₁
  ... | Just x rewrite x = Just (trans d d₁)
  shorten-der1 w (cong {w₁} {w'} {v} {v'} d d₁) with w =m? w' • v'
  ... | Just x rewrite x = Just refl
  ... | Nothing with w =m? w₁ • v
  ... | Just x rewrite x = Just (cong d d₁)
  ... | Nothing = Nothing
  shorten-der1 w (assoc {w₁} {v = v} {u}) with w =m? (w₁ • v • u)
  ... | Just x rewrite x = Just refl
  ... | Nothing with w =m? ((w₁ • v) • u)
  ... | Just x rewrite x = Just (assoc)
  ... | Nothing = Nothing
  shorten-der1 w (left-unit {v}) with w =m? (ε • v)
  ... | Just x rewrite x = Just (left-unit)
  ... | Nothing with w =m? v
  ... | Just x rewrite x = Just refl
  ... | Nothing = Nothing
  shorten-der1 {t = t} w (right-unit {v}) with w =m? (v • ε)
  ... | Just x rewrite x = Just (right-unit)
  ... | Nothing with w =m? v
  ... | Just x rewrite x = Just refl
  ... | Nothing = Nothing


  shorten-der1s : ∀ {X : Set} {Γ : Context X} {v t} -> {{meq : MaybeEq X}} -> (n : Nat) -> (w : Word X) -> Γ ⊢ v === t -> Maybe (Γ ⊢ w === t)
  shorten-der1s zero w d = Nothing
  shorten-der1s (succ n) w d with shorten-der1 w d
  shorten-der1s (succ n) w d | Just (trans s1 s2) with shorten-der1s n w s2
  shorten-der1s (succ n) w d | Just (trans s1 s2) | Just x = Just x
  shorten-der1s (succ n) w d | Just (trans s1 s2) | Nothing = Just (trans s1 s2)
  shorten-der1s (succ n) w d | Just x = Just x
  shorten-der1s (succ n) w d | Nothing = Nothing


  shorten-der : ∀ {X : Set} {Γ : Context X} {s t} -> {{meq : MaybeEq X}} -> (n : Nat) -> Γ ⊢ s === t -> (Γ ⊢ s === t)
  shorten-der zero d = d
  shorten-der {s = s} {t} (succ n) (trans {v = v} d d₁) with shorten-der1s (succ n) v d₁
  shorten-der {s = s} {t} (succ n) (trans {v = v} d d₁) | Just x with shorten-der n d 
  shorten-der {s = s} {t} (succ n) (trans {v = v} d d₁) | Just x | refl = (shorten-der n (flatten-trans 10000 x))
  shorten-der {s = s} {t} (succ n) (trans {v = v} d d₁) | Just x | d' = trans d' (shorten-der n (flatten-trans 10000 x))
  shorten-der {s = s} {t} (succ n) (trans {v = v} d d₁) | Nothing with shorten-der n d
  shorten-der {s = s} {t} (succ n) (trans {v = v} d d₁) | Nothing | refl = (shorten-der n d₁)
  shorten-der {s = s} {t} (succ n) (trans {v = v} d d₁) | Nothing | d' = trans d' (shorten-der n d₁)
  -- testing
  shorten-der n (cong d1 d2) = cong (shorten-der n d1) (shorten-der n d2)
  shorten-der n d = d

  simplify-der : ∀ {X : Set} {Γ : Context X} {l r} -> {{meq : MaybeEq X}} -> (n : Nat) -> Γ ⊢ l === r -> Γ ⊢ l === r
  simplify-der zero d = d
  simplify-der n d = shorten-der n (flatten-trans 10000 (flatten-cong 10 (flatten-symm 10000 d )))


-- ----------------------------------------------------------------------
-- * Basic lemmas about monoids

-- We take X and Γ to be fixed throughout the following lemmas.
module Monoid-Lemmas {X : Set} {Γ : Context X} where

  open Monoid-Equational

  -- The left unit is unique.
  lemma-left-unit-unique : {w : Word X} -> (∀ v -> Γ ⊢ w • v === v) -> Γ ⊢ w === ε
  lemma-left-unit-unique {w} p =
    equational w
            by right-unit reversed
        equals w • ε
            by p ε
        equals ε

  -- The right unit is unique.
  lemma-right-unit-unique : {w : Word X} -> (∀ v -> Γ ⊢ v • w === v) -> Γ ⊢ w === ε
  lemma-right-unit-unique {w} p =
    equational w
            by left-unit reversed
        equals ε • w
            by p ε
        equals ε

  -- Congruence for powers.
  lemma-cong-power : {w v : Word X} (n : Nat) -> Γ ⊢ w === v -> Γ ⊢ w ^ n === v ^ n
  lemma-cong-power zero hyp = refl
  lemma-cong-power (succ zero) hyp = hyp
  lemma-cong-power (succ (succ n)) hyp = cong hyp (lemma-cong-power (succ n) hyp)

  -- Recursive expansion of powers by repeated multiplication on the left.
  lemma-power-succ-left : {w : Word X} (n : Nat) -> Γ ⊢ w ^ (succ n) === w • w ^ n
  lemma-power-succ-left zero = right-unit reversed
  lemma-power-succ-left (succ n) = refl

  -- Recursive expansion of powers by repeated multiplication on the right.
  lemma-power-succ-right : {w : Word X} (n : Nat) -> Γ ⊢ w ^ (succ n) === w ^ n • w
  lemma-power-succ-right zero = left-unit reversed
  lemma-power-succ-right {w} (succ n) =
      equational w • w ^ succ n
              by right lemma-power-succ-right n
          equals w • (w ^ n • w)
              by assoc reversed
          equals (w • w ^ n) • w
              by left lemma-power-succ-left n reversed
          equals w ^ (succ n) • w

  -- A lemma about powers and multiplication.
  lemma-comm-power : ∀ {v w' w} -> (n : Nat) -> Γ ⊢ v • w === w' • v -> Γ ⊢ v • w ^ n === w' ^ n • v
  lemma-comm-power {v} {w'} {w} zero hyp =
      equational v • ε
              by right-unit
          equals v
              by left-unit reversed
          equals ε • v
  lemma-comm-power {v} {w'} {w} (succ n) hyp =
      equational v • w ^ succ n
              by right lemma-power-succ-left n
          equals v • (w • w ^ n)
              by assoc reversed
          equals (v • w) • w ^ n
              by left hyp
          equals (w' • v) • w ^ n
              by assoc
          equals w' • (v • w ^ n)
              by right lemma-comm-power n hyp
          equals w' • (w' ^ n • v)
              by assoc reversed
          equals (w' • w' ^ n) • v
              by left lemma-power-succ-left n reversed
          equals w' ^ succ n • v

  -- Commutativity of powers.
  lemma-comm-powers : ∀ {w v} -> (n m : Nat) -> Γ ⊢ w • v === v • w -> Γ ⊢ w ^ n • v ^ m === v ^ m • w ^ n
  lemma-comm-powers n m hyp = lemma-comm-power n (lemma-comm-power m hyp reversed) reversed

  -- Powers of products.
  lemma-product-power : ∀ {w v} -> (n : Nat) -> Γ ⊢ w • v === v • w -> Γ ⊢ (w • v) ^ n === w ^ n • v ^ n
  lemma-product-power zero hyp = right-unit reversed
  lemma-product-power {w} {v} (succ n) hyp =
      equational (w • v) ^ succ n
              by lemma-power-succ-left n
          equals (w • v) • (w • v) ^ n
              by right lemma-product-power n hyp
          equals (w • v) • (w ^ n • v ^ n)
              by assoc
          equals w • (v • (w ^ n • v ^ n))
              by right assoc reversed
          equals w • (v • w ^ n) • v ^ n
              by right left lemma-comm-power n (hyp reversed)
          equals w • (w ^ n • v) • v ^ n
              by right assoc
          equals w • (w ^ n • (v • v ^ n))
              by assoc reversed
          equals (w • w ^ n) • (v • v ^ n)
              by cong (lemma-power-succ-left n) (lemma-power-succ-left n) reversed
          equals w ^ succ n • v ^ succ n
  

  -- Power and change of basis.
  lemma-basis-change : ∀ {w w' v} -> (n : Nat) -> Γ ⊢ w' • w === ε -> Γ ⊢ (w • v • w') ^ succ n === w • v ^ succ n • w'
  lemma-basis-change {w} {w'} {v} 0 hyp = refl
  lemma-basis-change {w} {w'} {v} (succ n) hyp =
    equational (w • v • w') ^ succ (succ n)
            by definition
        equals (w • v • w') • (w • v • w') ^ succ n
            by right (lemma-basis-change n hyp)
        equals (w • v • w') • w • v ^ succ n • w'
            by left assoc reversed
        equals ((w • v) • w') • w • v ^ succ n • w'
            by assoc reversed
        equals (((w • v) • w') • w) • v ^ succ n • w'
            by left assoc
        equals ((w • v) • (w' • w)) • v ^ succ n • w'
            by left right hyp
        equals ((w • v) • ε) • v ^ succ n • w'
            by left right-unit
        equals (w • v) • v ^ succ n • w'
            by assoc
        equals w • v • v ^ succ n • w'
            by right assoc reversed
        equals w • (v • v ^ succ n) • w'
            by definition
        equals w • v ^ succ (succ n) • w'

-- ----------------------------------------------------------------------
-- ** Basic lemmas about groups

-- A context is group-like if it provides a left inverse to every
-- generator.
Grouplike : {X : Set} -> (Γ : Context X) -> Set
Grouplike {X} Γ = ∀ (x : X) -> ∃ λ (x' : Word X) -> Γ ⊢ x' • gen x === ε

-- We take X and Γ to be fixed throughout the following lemmas, and we
-- assume Γ is group-like.
module Group-Lemmas
       (X : Set)
       (Γ : Context X)
       (group-like : Grouplike Γ)
  where

  open Monoid-Equational

  infix 8 _⁻¹

  -- The inverse of a word.
  _⁻¹ : Word X -> Word X
  gen x ⁻¹ = fst (group-like x)
  ε ⁻¹ = ε
  (u • v) ⁻¹ = v ⁻¹ • u ⁻¹

  -- g ⁻¹ is a left inverse.
  lemma-left-inverse : {g : Word X} -> Γ ⊢ g ⁻¹ • g === ε
  lemma-left-inverse {gen x} = snd (group-like x)
  lemma-left-inverse {ε} = left-unit
  lemma-left-inverse {u • v} =
         equational (v ⁻¹ • u ⁻¹) • (u • v)
                 by assoc
             equals v ⁻¹ • (u ⁻¹ • (u • v))
                 by right assoc reversed
             equals v ⁻¹ • ((u ⁻¹ • u) • v)
                 by right left lemma-left-inverse
             equals v ⁻¹ • ε • v
                 by right left-unit
             equals v ⁻¹ • v
                 by lemma-left-inverse
             equals ε

  -- g ⁻¹ is a right inverse.
  lemma-right-inverse : {g : Word X} -> Γ ⊢ g • g ⁻¹ === ε
  lemma-right-inverse {g} = 
    equational g • (g ⁻¹)
            by left-unit reversed
        equals ε • (g • (g ⁻¹))
            by left lemma-left-inverse reversed
        equals ((g ⁻¹) ⁻¹ • g ⁻¹) • (g • (g ⁻¹))
            by assoc
        equals (g ⁻¹) ⁻¹ • (g ⁻¹ • (g • (g ⁻¹)))
            by right assoc reversed
        equals (g ⁻¹) ⁻¹ • ((g ⁻¹ • g) • g ⁻¹)
            by right left lemma-left-inverse
        equals (g ⁻¹) ⁻¹ • (ε • g ⁻¹)
            by right left-unit
        equals (g ⁻¹) ⁻¹ • g ⁻¹
            by lemma-left-inverse
        equals ε
  
  -- Left cancellation.
  lemma-left-cancel : {g h h' : Word X} -> Γ ⊢ g • h === g • h' -> Γ ⊢ h === h'
  lemma-left-cancel {g} {h} {h'} p = 
    equational h
               by left-unit reversed
        equals ε • h
               by left lemma-left-inverse reversed
        equals (g ⁻¹ • g) • h
               by assoc
        equals g ⁻¹ • (g • h)
               by right p
        equals g ⁻¹ • (g • h')
               by assoc reversed
        equals (g ⁻¹ • g) • h'
               by left lemma-left-inverse
        equals ε • h'
               by left-unit
        equals h'
  
  -- Right cancellation.
  lemma-right-cancel : {g g' h : Word X} -> Γ ⊢ g • h === g' • h -> Γ ⊢ g === g'
  lemma-right-cancel {g} {g'} {h} p = 
    equational g
               by right-unit reversed
        equals g • ε
               by right lemma-right-inverse reversed
        equals g • (h • h ⁻¹)
               by assoc reversed
        equals (g • h) • h ⁻¹
               by left p
        equals (g' • h) • h ⁻¹
               by assoc
        equals g' • (h • h ⁻¹)
               by right lemma-right-inverse
        equals g' • ε
               by right-unit
        equals g'
  
  -- Left inverses are unique.
  lemma-left-inverse-unique : {g h : Word X} -> Γ ⊢ h • g === ε -> Γ ⊢ h === g ⁻¹
  lemma-left-inverse-unique {g} {h} p = 
    equational h
               by right-unit reversed
        equals h • ε
               by right lemma-right-inverse reversed
        equals h • (g • g ⁻¹)
               by assoc reversed
        equals (h • g) • g ⁻¹
               by left p
        equals ε • g ⁻¹
               by left-unit
        equals g ⁻¹
  
  -- Right inverses are unique.
  lemma-right-inverse-unique : {g h : Word X} -> Γ ⊢ g • h === ε -> Γ ⊢ h === g ⁻¹
  lemma-right-inverse-unique {g} {h} p = 
    equational h
               by left-unit reversed
        equals ε • h
               by left lemma-left-inverse reversed
        equals (g ⁻¹ • g) • h
               by assoc
        equals g ⁻¹ • (g • h)
               by right p
        equals g ⁻¹ • ε
               by right-unit
        equals g ⁻¹

  -- Congruence for inverses.
  lemma-cong-inv : {g h : Word X} -> Γ ⊢ g === h -> Γ ⊢ g ⁻¹ === h ⁻¹
  lemma-cong-inv {g} {h} p = lemma-right-inverse-unique claim
     where
       claim : Γ ⊢ h • g ⁻¹ === ε
       claim = equational h • g ⁻¹
                       by left p reversed
                   equals g • g ⁻¹
                       by lemma-right-inverse
                   equals ε

  -- The inverse is involutive.
  lemma-inverse-involutive : {g : Word X} -> Γ ⊢ (g ⁻¹) ⁻¹ === g
  lemma-inverse-involutive {g} = lemma-right-inverse-unique lemma-left-inverse reversed
  
  -- The inverse of ε.
  lemma-unit-inverse : Γ ⊢ ε ⁻¹ === ε
  lemma-unit-inverse = refl
  
  -- The inverse of a product.
  lemma-product-inverse : ∀ {g h : Word X} -> Γ ⊢ (g • h) ⁻¹ === h ⁻¹ • g ⁻¹
  lemma-product-inverse = refl

  -- To show two words are equal, it is sufficient to show their
  -- inverses are equal.
  lemma-rule-inverse : ∀ {u v : Word X} -> Γ ⊢ u ⁻¹ === v ⁻¹ -> Γ ⊢ u === v
  lemma-rule-inverse {u} {v} hyp =
         equational u
                 by right-unit reversed
             equals u • ε
                 by right (lemma-left-inverse reversed)
             equals u • (v ⁻¹ • v)
                 by assoc reversed
             equals (u • v ⁻¹) • v
                 by left (right (hyp reversed))
             equals (u • u ⁻¹) • v
                 by left lemma-right-inverse
             equals ε • v
                 by left-unit
             equals v

  -- Commutativity of inverses. (Note: commutativity is the special
  -- case where v = v' and w = w')
  lemma-comm-inv : ∀ {v v' w w'} -> Γ ⊢ v • w === w' • v' -> Γ ⊢ v' • w ⁻¹ === w' ⁻¹ • v
  lemma-comm-inv {v} {v'} {w} {w'} hyp =
      equational v' • w ⁻¹
              by left-unit reversed
          equals ε • (v' • w ⁻¹)
              by left lemma-left-inverse reversed
          equals (w' ⁻¹ • w') • (v' • w ⁻¹)
              by assoc
          equals w' ⁻¹ • (w' • (v' • w ⁻¹))
              by right assoc reversed
          equals w' ⁻¹ • (w' • v') • w ⁻¹
              by right left hyp reversed
          equals w' ⁻¹ • (v • w) • w ⁻¹
              by right assoc
          equals w' ⁻¹ • v • (w • w ⁻¹)
              by assoc reversed
          equals (w' ⁻¹ • v) • (w • w ⁻¹)
              by right lemma-right-inverse
          equals (w' ⁻¹ • v) • ε
              by right-unit
          equals w' ⁻¹ • v

  -- Any equation can be reduced to a one-sided equation.
  lemma-one-sided : ∀ {w u} -> Γ ⊢ w • u ⁻¹ === ε -> Γ ⊢ w === u
  lemma-one-sided {w} {u} hyp =
      equational w
              by right-unit reversed
          equals w • ε
              by right lemma-left-inverse reversed
          equals w • (u ⁻¹ • u)
              by assoc reversed
          equals (w • u ⁻¹) • u
              by left hyp
          equals ε • u
              by left-unit
          equals u

-- ----------------------------------------------------------------------
-- * Subtheories

module Monoid-Subtheories where

  -- Γ is a subtheory of Γ' if each theorem of Γ is a theorem of Γ'.
  _⊆_ : {X : Set} -> (Γ Γ' : Context X) -> Set
  Γ ⊆ Γ' = ∀ {E} -> Γ ⊢ E -> Γ' ⊢ E
  
  -- Lemma: To show Γ ⊆ Γ', it suffices to show that each axiom of Γ
  -- is a theorem of Γ'.
  lemma-⊆ : ∀ {X : Set} {Γ Γ'} -> (∀ {E : Relation X} -> E ∈ Γ -> Γ' ⊢ E) -> Γ ⊆ Γ'
  lemma-⊆ hyp (axiom x) = hyp x
  lemma-⊆ hyp refl = refl
  lemma-⊆ hyp (symm p) = symm (lemma-⊆ hyp p)
  lemma-⊆ hyp (trans p q) = trans (lemma-⊆ hyp p) (lemma-⊆ hyp q)
  lemma-⊆ hyp (cong p q) = cong (lemma-⊆ hyp p) (lemma-⊆ hyp q)
  lemma-⊆ hyp assoc = assoc
  lemma-⊆ hyp left-unit = left-unit
  lemma-⊆ hyp right-unit = right-unit

